# Contains code from opt_einsum, which is licensed under the MIT License.
# The MIT License (MIT)

# Copyright (c) 2014 Daniel Smith

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

import numpy as np
import pytest

import pytblis


def random_scalar(is_complex, rng=None):
    if rng is None:
        rng = np.random.default_rng(0)
    if is_complex:
        return rng.random() + 1j * rng.random()
    return rng.random()


def random_array(shape, scalar_type, rng=None):
    if rng is None:
        rng = np.random.default_rng(0)
    if np.issubdtype(scalar_type, np.complexfloating):
        arr = rng.random(shape) + 1j * rng.random(shape)
    else:
        arr = rng.random(shape)
    return arr.astype(scalar_type)


def test_pytblis_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "pytblis" in sys.modules


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
def test_transpose_add_size0(scalar_type):
    rng = np.random.default_rng(0)
    A = random_array((0, 3, 3), scalar_type, rng=rng)
    B = random_array((3, 3), scalar_type, rng=rng)
    C = pytblis.transpose_add("Zij->ij", A, alpha=1.0)
    C_correct = np.einsum("Zij->ij", A)
    assert np.allclose(C, C_correct)

    A = random_array((0, 3, 3), scalar_type, rng=rng)
    B = random_array((3, 3), scalar_type, rng=rng)
    C_correct = np.einsum("Zij->ij", A) + 0.5 * B
    C = pytblis.transpose_add("Zij->ij", A, alpha=1.0, beta=0.5, out=B)
    assert np.allclose(C, C_correct)


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
def test_tensordot(scalar_type):
    rng = np.random.default_rng(0)
    A = random_array((3, 3, 3), scalar_type, rng=rng)
    B = random_array((3, 3, 3), scalar_type, rng=rng)

    C = pytblis.tensordot(A, B, axes=([2], [0]))
    C_correct = np.tensordot(A, B, axes=([2], [0]))
    assert np.allclose(C, C_correct)

    A = rng.random((3, 5, 4)).astype(scalar_type)
    B = rng.random((5, 4, 3)).astype(scalar_type)

    if np.iscomplexobj(A) or np.iscomplexobj(B):
        A = A + 1j * rng.random((3, 5, 4)).astype(scalar_type)
        B = B + 1j * rng.random((5, 4, 3)).astype(scalar_type)

    C = pytblis.tensordot(A, B, axes=2)
    C_correct = np.tensordot(A, B, axes=2)
    assert np.allclose(C, C_correct)


def test_tensordot_type_mixed():
    rng = np.random.default_rng(0)
    A = rng.random((3, 3)).astype(np.float32)
    B = rng.random((3, 3)).astype(np.float64)

    with pytest.warns(
        UserWarning,
        match="The types of the input tensors do not match. Falling back to numpy tensordot.",
    ):
        C = pytblis.tensordot(A, B, axes=([0], [0]))

    C_correct = np.tensordot(A, B, axes=([0], [0]))
    assert np.allclose(C, C_correct)


def test_tensordot_type_unsupported():
    rng = np.random.default_rng(0)
    A = rng.random((3, 3)).astype(np.int32)
    B = rng.random((3, 3)).astype(np.int32)

    with pytest.warns(
        UserWarning,
        match="TBLIS only supports float32, float64, complex64, and complex128. Falling back to numpy tensordot.",
    ):
        C = pytblis.tensordot(A, B, axes=([0], [0]))

    C_correct = np.tensordot(A, B, axes=([0], [0]))
    assert np.allclose(C, C_correct)


tests = [
    # Test scalar-like operations
    "a,->a",
    "ab,->ab",
    ",ab,->ab",
    ",,->",
    # Test hadamard-like products
    "a,ab,abc->abc",
    "a,b,ab->ab",
    # Test index-transformations
    "ea,fb,gc,hd,abcd->efgh",
    "ea,fb,abcd,gc,hd->efgh",
    "abcd,ea,fb,gc,hd->efgh",
    # Test complex contractions
    "acdf,jbje,gihb,hfac,gfac,gifabc,hfac",
    "acdf,jbje,gihb,hfac,gfac,gifabc,hfac",
    "cd,bdhe,aidb,hgca,gc,hgibcd,hgac",
    "abhe,hidj,jgba,hiab,gab",
    "bde,cdh,agdb,hica,ibd,hgicd,hiac",
    "chd,bde,agbc,hiad,hgc,hgi,hiad",
    "chd,bde,agbc,hiad,bdi,cgh,agdb",
    "bdhe,acad,hiab,agac,hibd",
    # Test collapse
    "ab,ab,c->",
    "ab,ab,c->c",
    "ab,ab,cd,cd->",
    "ab,ab,cd,cd->ac",
    "ab,ab,cd,cd->cd",
    "ab,ab,cd,cd,ef,ef->",
    # Test outer products
    "ab,cd,ef->abcdef",
    "ab,cd,ef->acdf",
    "ab,cd,de->abcde",
    "ab,cd,de->be",
    "ab,bcd,cd->abcd",
    "ab,bcd,cd->abd",
    # Random test cases that have previously failed
    "eb,cb,fb->cef",
    "dd,fb,be,cdb->cef",
    "bca,cdb,dbf,afc->",
    "dcc,fce,ea,dbf->ab",
    "fdf,cdd,ccd,afe->ae",
    "abcd,ad",
    "ed,fcd,ff,bcf->be",
    "baa,dcf,af,cde->be",
    "bd,db,eac->ace",
    "fff,fae,bef,def->abd",
    "efc,dbc,acf,fd->abe",
    # Inner products
    "ab,ab",
    "ab,ba",
    "abc,abc",
    "abc,bac",
    "abc,cba",
    #    "Za,Za",
    # GEMM test cases
    "ab,bc",
    "ab,cb",
    "ba,bc",
    "ba,cb",
    "abcd,cd",
    "abcd,ab",
    "abcd,cdef",
    "abcd,cdef->feba",
    "abcd,efdc",
    "Za,ab->ab",
    # Inner than dot
    "aab,bc ->ac",
    "ab,bcc->ac",
    "aab,bcc->ac",
    "baa,bcc->ac",
    "aab,ccb->ac",
    # Randomly build test caes
    "aab,fa,df,ecc->bde",
    "ecb,fef,bad,ed->ac",
    "bcf,bbb,fbf,fc->",
    "bb,ff,be->e",
    "bcb,bb,fc,fff->",
    "fbb,dfd,fc,fc->",
    "afd,ba,cc,dc->bf",
    "adb,bc,fa,cfc->d",
    "bbd,bda,fc,db->acf",
    "dba,ead,cad->bce",
    "aef,fbc,dca->bde",
    # single array transpose
    "ea->ea",
    "fb->fb",
    "abcd->dcab",
    "gc->cg",
    "hd->dh",
    "efgh->hfge",
    "acdf->afcd",
    "gihb->ghib",
    "hfac->cfah",
    "gfac->cgaf",
    "gifabc->abifcg",
    "hfac->cfha",
    # single array partial trace
    "gfac->caf",
    "gcac->ca",
]

_sizes = [2, 3, 4, 5, 4, 3, 2, 6, 5, 4, 3, 2, 5, 7, 4, 3, 2, 3, 4, 9, 0, 2, 4, 5, 3, 2, 6]
_no_collision_chars = "".join(chr(i) for i in range(7000, 7007))
_valid_chars = "abcdefghijklmnopqABCZ" + _no_collision_chars
_default_dim_dict = dict(zip(_valid_chars, _sizes))


def build_shapes(string, dimension_dict=None):
    if dimension_dict is None:
        dimension_dict = _default_dim_dict

    shapes = []
    string = string.replace(" ", "")
    terms = string.split("->")[0].split(",")
    for term in terms:
        dims = [dimension_dict[x] for x in term]
        shapes.append(tuple(dims))
    return tuple(shapes)


def build_views(string, dimension_dict=None, dtype=np.float64, rng=None):
    if rng is None:
        rng = np.random.default_rng(0)
    views = []
    for shape in build_shapes(string, dimension_dict=dimension_dict):
        if shape:
            arr = rng.random(shape).astype(dtype)
            if np.iscomplexobj(arr):
                arr += 1j * rng.random(shape).astype(dtype)
            views.append(arr)
        else:
            views.append(dtype(rng.random()))
    return tuple(views)


def build_views_some_complex(string, dimension_dict=None, dtype=np.float64, rng=None, complex_letter="g"):
    if rng is None:
        rng = np.random.default_rng(0)
    views = []
    if dimension_dict is None:
        dimension_dict = _default_dim_dict

    shapes = []
    string = string.replace(" ", "")
    terms = string.split("->")[0].split(",")
    for term in terms:
        dims = [dimension_dict[x] for x in term]
        shapes.append(tuple(dims))

    for shape, term in zip(shapes, terms):
        if shape:
            arr = rng.random(shape).astype(dtype)
            if complex_letter in term:
                arr = arr + 1j * rng.random(shape).astype(dtype)
            views.append(arr)
        else:
            views.append(dtype(rng.random()))
    return tuple(views)


def build_views_multi_type(string, dimension_dict=None, dtypes=None, rng=None):
    if rng is None:
        rng = np.random.default_rng(0)
    views = []
    for shape, dtype in zip(build_shapes(string, dimension_dict=dimension_dict), dtypes):
        if shape:
            arr = rng.random(shape).astype(dtype)
            if np.iscomplexobj(arr):
                arr += 1j * rng.random(shape).astype(dtype)
            views.append(arr)
        else:
            views.append(dtype(rng.random()))
    return tuple(views)


@pytest.mark.parametrize("string", tests)
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_einsum(string, dtype):
    views = build_views(string, dtype=dtype)
    tblis_result = pytblis.einsum(string, *views)
    numpy_result = np.einsum(string, *views)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize("string", tests)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_einsum_complex_real(string, dtype):
    views = build_views_some_complex(string, dtype=dtype)
    tblis_result = pytblis.einsum(string, *views, complex_real_contractions=True)
    numpy_result = np.einsum(string, *views)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


single_array_tests = ["eZ", "fb", "abcd", "gc", "hd", "efgh", "acZf", "gihb", "hfac", "gfac", "gifabc", "hfac", "Z"]


@pytest.mark.parametrize("string", single_array_tests)
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_ascontiguousarray(string, dtype):
    rng = np.random.default_rng(0)
    views = build_views(string, dtype=dtype)
    arr = views[0]

    arr = np.transpose(arr, axes=rng.permutation(len(arr.shape)))
    tblis_result = pytblis.ascontiguousarray(arr)
    numpy_result = np.ascontiguousarray(arr)
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"
    assert tblis_result.strides == numpy_result.strides, f"Strides mismatch for string: {string}"


@pytest.mark.parametrize("string", single_array_tests)
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_asfortranarray(string, dtype):
    rng = np.random.default_rng(0)
    views = build_views(string, dtype=dtype)
    arr = views[0]

    arr = np.transpose(arr, axes=rng.permutation(len(arr.shape)))
    tblis_result = pytblis.asfortranarray(arr)
    numpy_result = np.asfortranarray(arr)
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"
    assert tblis_result.strides == numpy_result.strides, f"Strides mismatch for string: {string}"


@pytest.mark.parametrize("string", single_array_tests)
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_tensor_transpose(string, dtype):
    rng = np.random.default_rng(0)
    views = build_views(string, dtype=dtype)
    arr = views[0]

    perm = rng.permutation(len(arr.shape))

    string_perm = "".join(np.array(list(string))[perm])
    command_string = f"{string}->{string_perm}"

    numpy_result = np.transpose(arr, axes=perm)
    tblis_result = pytblis.transpose_add(command_string, arr, alpha=1.0)
    assert np.allclose(tblis_result, numpy_result), f"Failed for command: {command_string}"


@pytest.mark.parametrize("string", single_array_tests)
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_tensor_transpose_add(string, dtype):
    rng = np.random.default_rng(0)
    a = build_views(string, dtype=dtype, rng=rng)[0]

    alpha = random_scalar(np.iscomplexobj(a), rng=rng)
    beta = random_scalar(np.iscomplexobj(a), rng=rng)
    perm = rng.permutation(len(a.shape))

    b = build_views(string, dtype=dtype, rng=rng)[0]
    b = np.ascontiguousarray(np.transpose(b, axes=perm))

    string_perm = "".join(np.array(list(string))[perm])
    command_string = f"{string}->{string_perm}"

    numpy_result = beta * b + alpha * np.transpose(a, axes=perm)
    tblis_result = pytblis.transpose_add(command_string, a, alpha=alpha, beta=beta, out=b)
    assert np.allclose(tblis_result, numpy_result), f"Failed for command: {command_string}"


contraction_tests = [
    "ijk, jkl->il",
    "abZd, Zd->ab",
    "aab, bZ->aZ",
    "ab, ba->",
    "abc, cba->",
    "abcd, cdef->abef",
    "aab, bcc->ac",
    "bbZZ, aaZ->baZ",
]


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("string", contraction_tests)
def test_contract(string, scalar_type):
    rng = np.random.default_rng(0)
    a, b = build_views(string, dtype=scalar_type, rng=rng)
    tblis_result = pytblis.contract(string, a, b)
    numpy_result = np.einsum(string, a, b)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("string", contraction_tests)
def test_contract_with_out(string, scalar_type):
    rng = np.random.default_rng(0)
    a, b = build_views(string, dtype=scalar_type, rng=rng)
    numpy_result = np.einsum(string, a, b)
    out = np.empty_like(numpy_result)
    tblis_result = pytblis.contract(string, a, b, out=out)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("string", contraction_tests)
def test_contract_alpha(string, scalar_type):
    rng = np.random.default_rng(0)
    alpha = random_scalar(np.iscomplexobj(scalar_type()), rng=rng)
    a, b = build_views(string, dtype=scalar_type, rng=rng)
    tblis_result = pytblis.contract(string, a, b, alpha=alpha)
    numpy_result = alpha * np.einsum(string, a, b)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("string", contraction_tests)
def test_contract_with_out_alphabeta(string, scalar_type):
    rng = np.random.default_rng(0)
    alpha = random_scalar(np.iscomplexobj(scalar_type()), rng=rng)
    beta = random_scalar(np.iscomplexobj(scalar_type()), rng=rng)
    a, b = build_views(string, dtype=scalar_type, rng=rng)
    numpy_result = alpha * np.einsum(string, a, b)
    C = random_array(numpy_result.shape, scalar_type, rng=rng)
    numpy_result += beta * C
    tblis_result = pytblis.contract(string, a, b, alpha=alpha, beta=beta, out=C)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize(
    "scalar_types",
    [
        (np.float32, np.complex64),
        (np.float64, np.complex128),
        (np.complex64, np.float32),
        (np.complex128, np.float64),
    ],
)
@pytest.mark.parametrize("string", contraction_tests)
@pytest.mark.parametrize("conja", [False, True])
@pytest.mark.parametrize("conjb", [False, True])
def test_contract_mixedtype(string, scalar_types, conja, conjb):
    rng = np.random.default_rng(0)

    alpha = random_scalar(False, rng=rng)
    a, b = build_views_multi_type(string, dtypes=scalar_types, rng=rng)
    aifconj = a.conj() if conja else a
    bifconj = b.conj() if conjb else b
    numpy_result = alpha * np.einsum(string, aifconj, bifconj)
    tblis_result = pytblis.contract(string, a, b, alpha=alpha, conja=conja, conjb=conjb, complex_real_contractions=True)
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize("scalar_type1", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("scalar_type2", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("string", contraction_tests)
@pytest.mark.parametrize("conja", [False, True])
@pytest.mark.parametrize("conjb", [False, True])
def test_contract_mixedtype_alphabeta(string, scalar_type1, scalar_type2, conja, conjb):
    rng = np.random.default_rng(0)
    scalar_types = (scalar_type1, scalar_type2)
    alpha = random_scalar(False, rng=rng)
    beta = random_scalar(False, rng=rng)
    a, b = build_views_multi_type(string, dtypes=scalar_types, rng=rng)
    aifconj = a.conj() if conja else a
    bifconj = b.conj() if conjb else b
    numpy_result = alpha * np.einsum(string, aifconj, bifconj)
    C = random_array(numpy_result.shape, np.result_type(*scalar_types), rng=rng)
    numpy_result += beta * C
    if a.real.dtype.type != b.real.dtype.type:
        with pytest.raises(TypeError):
            tblis_result = pytblis.contract(
                string, a, b, alpha=alpha, beta=beta, out=C, conja=conja, conjb=conjb, complex_real_contractions=True
            )
        return
    tblis_result = pytblis.contract(
        string, a, b, alpha=alpha, beta=beta, out=C, conja=conja, conjb=conjb, complex_real_contractions=True
    )
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"


@pytest.mark.parametrize("scalar_type", [np.float32, np.float64, np.complex64, np.complex128])
@pytest.mark.parametrize("string", single_array_tests)
@pytest.mark.parametrize("conj", [False, True])
def test_complexify(string, scalar_type, conj):
    rng = np.random.default_rng(0)
    a_real = build_views(string, dtype=scalar_type, rng=rng)[0]
    a_imag = build_views(string, dtype=scalar_type, rng=rng)[0]
    if scalar_type in [np.complex64, np.complex128]:
        with pytest.raises(AssertionError, match="Inputs must be real arrays."):
            pytblis.complexify(a_real, a_imag, conj=conj)
        return
    tblis_result = pytblis.complexify(a_real, a_imag, conj=conj)
    numpy_result = a_real + 1j * a_imag
    if conj:
        numpy_result = numpy_result.conj()
    assert tblis_result.shape == numpy_result.shape, f"Shape mismatch for string: {string}"
    assert np.allclose(tblis_result, numpy_result), f"Failed for string: {string}"



# Contains code from opt_einsum, which is licensed under the MIT License.
# The MIT License (MIT)

# Copyright (c) 2014 Daniel Smith

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np

from .wrappers import contract, transpose_add


def einsum(*operands, out=None, optimize="greedy", complex_real_contractions=False, **kwargs):
    """
    einsum(subscripts, *operands, out=None, order='K',
           optimize='greedy')
    Evaluates the Einstein summation convention on the operands.

    Drop-in replacement for numpy.einsum, using TBLIS for tensor contractions.

    Parameters
    ----------
    subscripts : str
        Specifies the subscripts for summation as comma separated list of
        subscript labels. An implicit (classical Einstein summation)
        calculation is performed unless the explicit indicator '->' is
        included as well as subscript labels of the precise output form.
    operands : list of array_like
        These are the arrays for the operation.
    out : ndarray, optional
        If provided, the calculation is done into this array.
    order : {'C', 'F', 'A', 'K'}, optional
        Controls the memory layout of the output. 'C' means it should
        be C contiguous. 'F' means it should be Fortran contiguous,
        'A' means it should be 'F' if the inputs are all 'F', 'C' otherwise.
        'K' means it should be as close to the layout as the inputs as
        is possible, including arbitrarily permuted axes.
        Default is 'K'.
    optimize : {'greedy', 'optimal'}, default 'greedy'
        Controls the optimization strategy used to compute the contraction.
    complex_real_contractions : bool, default False
        If True, handle contractions between complex and real tensors by performing
        separate contractions for the real and imaginary parts of the complex tensor.

    Returns
    -------
    output : ndarray
        The calculation based on the Einstein summation convention.
    """
    specified_out = out is not None
    assert optimize in ("greedy", "optimal"), "optimize must be 'greedy' or 'optimal'"

    # Check the kwargs to avoid a more cryptic error later, without having to
    # repeat default values here
    valid_einsum_kwargs = ["order"]
    unknown_kwargs = [k for (k, v) in kwargs.items() if k not in valid_einsum_kwargs]
    if unknown_kwargs:
        msg = f"Did not understand the following kwargs: {unknown_kwargs}"
        raise TypeError(msg)

    # Build the contraction list and operand
    operands, contraction_list = np.einsum_path(*operands, optimize=optimize, einsum_call=True)

    # Handle order kwarg for output array, c_einsum allows mixed case
    output_order = kwargs.pop("order", "K")
    if output_order.upper() == "A":
        output_order = "F" if all(arr.flags.f_contiguous for arr in operands) else "C"

    # Start contraction loop
    for num, contraction in enumerate(contraction_list):
        if len(contraction) == 3:  # numpy 2.4.0 and newer.
            inds, einsum_str, _ = contraction
        else:  # numpy 2.3.x and older.
            inds, _, einsum_str, _, _ = contraction
        tmp_operands = [operands.pop(x) for x in inds]
        # Do we need to deal with the output?
        handle_out = specified_out and ((num + 1) == len(contraction_list))
        out_kwarg = None
        if handle_out:
            out_kwarg = out

        if len(tmp_operands) == 2:
            new_view = contract(
                einsum_str,
                *tmp_operands,
                out=out_kwarg,
                allow_partial_trace=True,
                complex_real_contractions=complex_real_contractions,
            )

        elif len(tmp_operands) == 1:
            # check if only a transpose
            einsum_str = einsum_str.replace(" ", "")
            subscript_a, subscript_b = einsum_str.split("->")
            if sorted(subscript_a) == sorted(subscript_b):
                # only a transpose, use numpy for this (should return view)
                new_view = np.einsum(einsum_str, tmp_operands[0], out=out_kwarg, **kwargs)
            else:
                # may involve a trace or replication, use tblis transpose_add for this
                new_view = transpose_add(einsum_str, tmp_operands[0], out=out_kwarg, **kwargs)
        else:
            # fallback to numpy einsum
            # e.g. contractions of 3 tensors
            out_kwarg = None
            if handle_out:
                out_kwarg = out
            new_view = np.einsum(einsum_str, *tmp_operands, out=out_kwarg, **kwargs)

        # Append new items and dereference what we can
        operands.append(new_view)
        del tmp_operands, new_view

    if specified_out:
        return out
    return operands[0]



# Contains code from NumPy, which is licensed under the BSD 3-Clause License.
# Copyright (c) 2005-2025, NumPy Developers.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

#     * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.

#     * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.

#     * Neither the name of the NumPy Developers nor the names of any
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np

from ._pytblis_impl import mult
from .typecheck import _accepted_types, _valid_labels


def tensordot(a, b, axes=2):
    """
    Compute tensor dot product along specified axes using TBLIS.

    Given two tensors, `a` and `b`, and an array_like object containing
    two array_like objects, ``(a_axes, b_axes)``, sum the products of
    `a`'s and `b`'s elements (components) over the axes specified by
    ``a_axes`` and ``b_axes``. The third argument can be a single non-negative
    integer_like scalar, ``N``; if it is such, then the last ``N`` dimensions
    of `a` and the first ``N`` dimensions of `b` are summed over.

    Parameters
    ----------
    a, b : array_like
        Tensors to "dot".

    axes : int or (2,) array_like
        * integer_like
          If an int N, sum over the last N axes of `a` and the first N axes
          of `b` in order. The sizes of the corresponding axes must match.
        * (2,) array_like
          Or, a list of axes to be summed over, first sequence applying to `a`,
          second to `b`. Both elements array_like must be of the same length.

    Returns
    -------
    output : ndarray
        The tensor dot product of the input.
    """
    try:
        iter(axes)
    except Exception:
        axes_a = list(range(-axes, 0))
        axes_b = list(range(axes))
    else:
        axes_a, axes_b = axes
    try:
        na = len(axes_a)
        axes_a = list(axes_a)
    except TypeError:
        axes_a = [axes_a]
        na = 1
    try:
        nb = len(axes_b)
        axes_b = list(axes_b)
    except TypeError:
        axes_b = [axes_b]
        nb = 1

    a, b = np.asarray(a), np.asarray(b)
    as_ = a.shape
    nda = a.ndim
    bs = b.shape
    ndb = b.ndim
    equal = True
    if na != nb:
        equal = False
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                break
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb
    if not equal:
        msg = "shape-mismatch for sum"
        raise ValueError(msg)

    if a.dtype.type != b.dtype.type:
        warnings.warn("The types of the input tensors do not match. Falling back to numpy tensordot.", stacklevel=2)
        return np.tensordot(a, b, axes=axes)

    if a.dtype.type not in _accepted_types:
        warnings.warn(
            "TBLIS only supports float32, float64, complex64, and complex128. Falling back to numpy tensordot.",
            stacklevel=2,
        )
        return np.tensordot(a, b, axes=axes)

    # Move the axes to sum over to the end of "a"
    # and to the front of "b"
    notin_a = [k for k in range(nda) if k not in axes_a]
    olda = [as_[axis] for axis in notin_a]

    notin_b = [k for k in range(ndb) if k not in axes_b]
    oldb = [bs[axis] for axis in notin_b]

    outshape = olda + oldb

    if len(outshape) + len(axes_a) > len(_valid_labels):
        msg = f"Too many axes: maximum is {len(_valid_labels)}"
        raise ValueError(msg)

    labels_common = _valid_labels[: len(axes_a)]
    remaining_labels = _valid_labels[len(axes_a) :].copy()
    inds_A = []
    inds_B = []
    inds_C = []
    for i in range(nda):
        if i not in axes_a:
            newind = remaining_labels.pop(0)
            inds_A.append(newind)
            inds_C.append(newind)
        else:
            inds_A.append(labels_common[axes_a.index(i)])

    for i in range(ndb):
        if i not in axes_b:
            newind = remaining_labels.pop(0)
            inds_B.append(newind)
            inds_C.append(newind)
        else:
            inds_B.append(labels_common[axes_b.index(i)])
    inds_A = "".join(inds_A)
    inds_B = "".join(inds_B)
    inds_C = "".join(inds_C)

    restype = np.result_type(a, b)
    c = np.empty(outshape, dtype=restype)

    mult(a, b, c, inds_A, inds_B, inds_C)
    return c



from typing import Union

import numpy as np

from .typecheck import _check_strides, _check_tblis_types
from .wrappers import transpose_add

scalar = Union[float, complex]


def complexify(realpart, imagpart, conj=False, scale=1.0, out=None):
    """Returns a complex tensor formed from the given real and imaginary parts.

    Parameters
    ----------
    realpart : array_like
        Must be a real-valued array.
    imagpart : array_like
        Must be a real-valued array with the same shape as `realpart`.
    conj : bool, optional
        If True, return (realpart - 1j * imagpart), by default False
    scale : float, optional
        Multiply the result by a real scalar, by default 1.0
    out : array_like, optional
        Output array to store the result, by default None.
        If given, must be a complex array with the same shape as `realpart` and `imagpart`,
        and use the same precision as the inputs.
    Returns
    -------
    output : ndarray
        The complex tensor formed from the given real and imaginary parts.

    Examples
    --------
    >>> import numpy as np
    >>> from pytblis import complexify
    >>> re = np.random.rand(3, 4, 5)
    >>> im = np.random.rand(3, 4, 5)
    >>> arr = complexify(re, im)
    """
    _check_tblis_types(realpart, imagpart, out=out)
    assert not np.iscomplexobj(realpart), "Inputs must be real arrays."
    assert not np.iscomplexobj(imagpart), "Inputs must be real arrays."
    input_strides_ok, output_strides_ok = _check_strides(realpart, imagpart, out=out)
    is_trivial = realpart.size == 0 or imagpart.size == 0
    shape = realpart.shape
    ndim = realpart.ndim
    assert shape == imagpart.shape, "Input arrays must have the same shape."
    inds = "abcdefghijklmnopqrstuvwxyz"[:ndim]
    result_type = np.result_type(realpart.dtype, 1j).type
    if out is None:
        out = np.empty(shape, dtype=result_type)

    if is_trivial:
        return np.asarray(realpart) + np.asarray(imagpart) * 1j

    assert out.shape == shape, "Output array must have the same shape as input arrays."
    assert input_strides_ok, "Input arrays must not have negative strides."
    assert output_strides_ok, "Output array must not have negative strides."

    conj_fac = -1.0 if conj else 1.0
    transpose_add(inds + "->" + inds, realpart, out=out.real, alpha=scale)
    transpose_add(inds + "->" + inds, imagpart, out=out.imag, alpha=scale * conj_fac)
    return out



from ._pytblis_impl import add, dot, get_num_threads, mult, reduce, reduce_t, set_num_threads, shift
from .einsum_impl import einsum
from .mixedtype import complexify
from .tensordot_impl import tensordot
from .wrappers import ascontiguousarray, asfortranarray, contract, transpose_add

__all__ = [
    "add",
    "ascontiguousarray",
    "asfortranarray",
    "complexify",
    "contract",
    "dot",
    "einsum",
    "get_num_threads",
    "mult",
    "reduce",
    "reduce_t",
    "set_num_threads",
    "shift",
    "tensordot",
    "transpose_add",
]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os, subprocess
# import sys
# sys.path.insert(0, os.path.abspath('.'))

subprocess.call('doxygen', shell=True)

# -- Project information -----------------------------------------------------

project = 'MArray'
copyright = '2021, Devin Matthews'
author = 'Devin Matthews'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'breathe',
#    'exhale',
    'myst_parser'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Setup the breathe extension
breathe_projects = {
    "MArray": "doxygen/xml"
    }
breathe_default_project = "MArray"
breathe_default_members = ('members',)

# Setup the exhale extension
#exhale_args = {
    # These arguments are required
#    "containmentFolder":     "./api",
#    "rootFileName":          "api.rst",
#    "doxygenStripFromPath":  "..",
    # Heavily encouraged optional argument (see docs)
#    "rootFileTitle":         "API Reference",
    # Suggested optional arguments
#    "createTreeView":        True,
    # TIP: if using the sphinx-bootstrap-theme, you need
    # "treeViewIsBootstrap": True,
#    "exhaleExecutesDoxygen": True,
#    "exhaleUseDoxyfile": True
#}

# Tell sphinx what the primary language being documented is.
#primary_domain = 'cpp'

# Tell sphinx what the pygments highlight language should be.
#highlight_language = 'cpp'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['custom.css']



import numpy as np

_valid_labels = list("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
_accepted_types = (np.float32, np.float64, np.complex64, np.complex128)


def _check_strides(*tensors, out=None):
    """
    Check for negative strides in the input tensors.
    Return False if any tensor has a negative stride, otherwise True.
    Non-contiguity is OK.
    """
    inputs_ok = all(all(s > 0 for s in tensor.strides) for tensor in tensors)
    output_ok = (out is None) or all(s > 0 for s in out.strides)
    return inputs_ok, output_ok


def _check_tblis_types(*tensors, out=None):
    """
    Returns the scalar type if all tensors have the same datatype, and this datatype is
    one of the supported types (float, double, complex float, complex double).
    Otherwise return None.
    """
    if len(tensors) == 0:
        return None
    first_type = tensors[0].dtype.type
    for tensor in tensors:
        if tensor.dtype.type != first_type or tensor.dtype.type not in _accepted_types:
            return None
    if out is not None and out.dtype.type != first_type:
        return None
    return first_type


def contraction_result_shape(subscripts, a_shape, b_shape):
    subscripts = subscripts.replace(" ", "")
    input_str, subscript_c = subscripts.split("->")
    subscript_a, subscript_b = input_str.split(",")
    if not (set(subscript_a) | set(subscript_b)) >= set(subscript_c):
        msg = f"Invalid subscripts '{subscripts}'"
        raise ValueError(msg)
    a_shape_dic = dict(zip(subscript_a, a_shape))
    b_shape_dic = dict(zip(subscript_b, b_shape))
    if any(a_shape_dic[x] != b_shape_dic[x] for x in set(subscript_a) & set(subscript_b)):
        msg = f"Shape mismatch for subscripts '{subscripts}': {a_shape} {b_shape}"
        raise ValueError(msg)

    ab_shape_dic = {**a_shape_dic, **b_shape_dic}
    return tuple(ab_shape_dic[x] for x in subscript_c)



import gdb
import re
 
class ShortVectorPrinter:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        size = int(self.val['_size'])
        N = int(self.val.type.template_argument(1))
        cap = N if size <= N else int(self.val['_capacity'])
        return 'MArray::short_vector<%d> of length %d, capacity %d' % (N, size, cap)

    def children(self):
        size = int(self.val['_size'])
        data = self.val['_alloc']['_data']
        for i in range(size):
            yield ('[%d]' % i, data.dereference())
            data = data + 1

    def display_hint(self):
        return 'array'
    
def str_lookup_function(val):
    lookup_tag = val.type.strip_typedefs().tag
    if lookup_tag == None:
        return None
    regex = re.compile("^MArray::short_vector<.*>$")
    if regex.match(lookup_tag):
        return ShortVectorPrinter(val)
    return None

gdb.pretty_printers.append(str_lookup_function)



import enum
from typing import Union

import numpy.typing as npt

scalar = Union[float, complex]

__all__ = ["add", "dot", "mult", "reduce", "shift"]

def add(
    a: npt.ArrayLike,
    b: npt.ArrayLike,
    idx_A: str,
    idx_B: str,
    alpha: scalar = 1.0,
    beta: scalar = 0.0,
    conja: bool = False,
    conjb: bool = False,
) -> None: ...
def dot(
    a: npt.ArrayLike,
    b: npt.ArrayLike,
    idx_A: str,
    idx_B: str,
    alpha: scalar = 1.0,
    beta: scalar = 1.0,
    conja: bool = False,
    conjb: bool = False,
) -> scalar: ...
def mult(
    a: npt.ArrayLike,
    b: npt.ArrayLike,
    c: npt.ArrayLike,
    idx_A: str,
    idx_B: str,
    idx_C: str,
    alpha: scalar = 1.0,
    beta: scalar = 0.0,
    conja: bool = False,
    conjb: bool = False,
) -> None: ...
def shift(
    a: npt.ArrayLike,
    idx_A: str,
    alpha: scalar = 1.0,
    beta: scalar = 0.0,
) -> None: ...
def tensor_set(
    a: npt.ArrayLike,
    idx_A: str,
    value: scalar,
) -> None: ...
def reduce(a: npt.ArrayLike, idx_A: str, op: enum.Enum, conja: bool = False) -> Union[scalar, tuple[scalar, int]]: ...
def get_num_threads() -> int: ...
def set_num_threads(num_threads: int) -> None: ...



import warnings
from typing import Optional, Union

import numpy as np
import numpy.typing as npt

from ._pytblis_impl import add, mult, shift
from .typecheck import _accepted_types, _check_strides, _check_tblis_types, _valid_labels, contraction_result_shape

scalar = Union[float, complex]


def transpose_add(
    subscripts: str,
    a: npt.ArrayLike,
    alpha: scalar = 1.0,
    beta: scalar = 0.0,
    out: Optional[npt.ArrayLike] = None,
    conja: bool = False,
    conjout: bool = False,
) -> npt.ArrayLike:
    """
    Perform tensor transpose and addition based on the provided subscripts.
    High-level wrapper for tblis_tensor_add.
    B (stored in `out` if provided) is computed as:
    B = alpha * transpose(a, subscripts) + beta * B.
    Optionally, conjugation can be applied to `a` and `out`.

    Parameters
    ----------
    subscripts : str
        Subscripts defining the contraction.
    a : array_like
        Tensor operand
    alpha : scalar, optional
        Scaling factor for `a`.
    beta : scalar, optional
        Scaling factor for the output tensor `b`. Must be 0.0 if `out` is None.
    conja: bool, optional
        If True, conjugate the first tensor `a`. Alpha is not conjugated.
    conjout: bool, optional
        If True, conjugate the output tensor `b`. Beta is not conjugated.
    out : array_like, optional
        Output tensor containing `b`.

    Returns
    -------
    ndarray
        Result of the tensor contraction.

    Examples
    --------
    >>> import numpy as np
    >>> from pytblis import transpose_add
    >>> a = np.random.rand(3, 4, 5)
    >>> b = transpose_add("ijk->ikj", a, alpha=1.0)
    """
    a = np.asarray(a)
    scalar_type = _check_tblis_types(a, out) if out is not None else _check_tblis_types(a)
    input_strides_ok, output_strides_ok = _check_strides(a, out=out)

    if scalar_type is None:
        raise TypeError(
            "TBLIS only supports float32, float64, complex64, and complex128. "
            "Types do not match or unsupported type detected. "
        )
    # It's okay if A has bad strides if its size is 0, because nothing will be read.
    if not input_strides_ok and a.size != 0:
        msg = f"Input tensor of shape {a.shape} has non-positive strides: {a.strides}"
        raise ValueError(msg)
    if not output_strides_ok and out.size != 0:
        msg = f"Output tensor of shape {out.shape} has non-positive strides: {out.strides}"
        raise ValueError(msg)

    subscripts = subscripts.replace(" ", "")
    a_idx, b_idx = subscripts.split("->")

    if not set(a_idx) >= set(b_idx):
        msg = f"Invalid subscripts '{subscripts}'"
        raise ValueError(msg)
    a_shape_dic = dict(zip(a_idx, a.shape))

    b_shape = tuple(a_shape_dic[x] for x in b_idx)

    if out is None:
        out = np.empty(b_shape, dtype=scalar_type)
        assert beta == 0.0, "beta must be 0.0 if out is None"
    else:
        out = np.asarray(out)
        if out.shape != b_shape:
            msg = f"Output shape {out.shape} does not match expected shape {b_shape} for subscripts '{subscripts}'"
            raise ValueError(msg)

    # handle zero-sized input
    # if a is size zero, tensor_add quits early and does not scale B.
    if a.size == 0:
        shift(out, b_idx, alpha=0.0, beta=beta)
    else:
        add(a, out, a_idx, b_idx, alpha=alpha, beta=beta, conja=conja, conjb=conjout)
    return out


def contract_same_type(
    subscripts: str,
    a: npt.ArrayLike,
    b: npt.ArrayLike,
    alpha: scalar = 1.0,
    beta: scalar = 0.0,
    out: Optional[npt.ArrayLike] = None,
    conja: bool = False,
    conjb: bool = False,
    allow_partial_trace: bool = False,
) -> npt.ArrayLike:
    """
    Perform tensor contraction based on the provided subscripts.
    C (stored in `out` if provided) is computed as:
    C = alpha * einsum(subscripts, a, b) + beta * C if `out` is provided.

    Parameters
    ----------
    subscripts : str
        Subscripts defining the contraction.
    a : array_like
        First tensor operand.
    b : array_like
        Second tensor operand.
    alpha : scalar, optional
        Scaling factor for the product of `a` and `b`.
    beta : scalar, optional
        Scaling factor for the output tensor. Must be 0.0 if `out` is None.
    conja: bool, optional
        If True, conjugate the first tensor `a` before contraction. Alpha is not conjugated.
    conjb: bool, optional
        If True, conjugate the second tensor `b` before contraction. Beta is not conjugated.
    allow_partial_trace : bool, optional
        If True, handle redundant indices in subscripts for `a` and `b` by doing partial trace before contraction.
    out : array_like, optional
        Output tensor to store the result.

    Returns
    -------
    ndarray
        Result of the tensor contraction.
    """
    a = np.asarray(a)
    b = np.asarray(b)
    scalar_type = _check_tblis_types(a, b, out=out)
    input_strides_ok, output_strides_ok = _check_strides(a, b, out=out)
    is_trivial = a.size == 0 or b.size == 0

    fallback = False

    if scalar_type is None:
        warnings.warn(
            "TBLIS only supports float32, float64, complex64, and complex128. "
            "Types do not match or unsupported type detected. "
            "Will attempt to fall back to numpy tensordot.",
            stacklevel=2,
        )
        fallback = True
    if not output_strides_ok and not is_trivial:
        warnings.warn(
            f"Output tensor of shape {out.shape} has non-positive strides: {out.strides}. "
            "Will attempt to fall back to numpy tensordot.",
            stacklevel=2,
        )
        fallback = True

    if not input_strides_ok and not is_trivial:
        warnings.warn(
            f"Input tensor of shape {a.shape} has non-positive strides: {a.strides}. "
            "Will attempt to fall back to numpy tensordot.",
            stacklevel=2,
        )
        fallback = True

    if fallback:
        if alpha != 1.0 or beta != 0.0:
            msg = "Cannot fall back to numpy tensordot unless alpha = 1.0 and beta = 0.0"
            raise ValueError(msg)
        return np.einsum(subscripts, a, b)

    subscripts = subscripts.replace(" ", "")
    input_str, subscript_c = subscripts.split("->")
    subscript_a, subscript_b = input_str.split(",")

    idx_a = frozenset(subscript_a)
    idx_b = frozenset(subscript_b)
    idx_c = frozenset(subscript_c)
    idx_redundant_a = idx_a - idx_b - idx_c
    idx_redundant_b = idx_b - idx_a - idx_c
    idx_redundant_c = idx_c - idx_a - idx_b

    if idx_redundant_c:
        raise RuntimeError("Should never have redundant indices in the output. This is probably a bug.")

    if not allow_partial_trace and (idx_redundant_a or idx_redundant_b):
        msg = (
            f"Subscripts '{subscripts}' require partial trace on "
            f"{'a ' if idx_redundant_a else ''}{'b' if idx_redundant_b else ''}. "
            "Pass allow_partial_trace=True to enable."
        )
        raise ValueError(msg)

    c_shape = contraction_result_shape(subscripts, a.shape, b.shape)

    if idx_redundant_a:
        # partial trace on a
        subscript_a_traced = "".join([i for i in subscript_a if i not in idx_redundant_a])
        einsum_str_traced = f"{subscript_a}->{subscript_a_traced}"
        a = transpose_add(einsum_str_traced, a)
        subscript_a = subscript_a_traced
    if idx_redundant_b:
        # partial trace on b
        subscript_b_traced = "".join([i for i in subscript_b if i not in idx_redundant_b])
        einsum_str_traced = f"{subscript_b}->{subscript_b_traced}"
        b = transpose_add(einsum_str_traced, b)
        subscript_b = subscript_b_traced

    if out is None:
        out = np.empty(c_shape, dtype=scalar_type)
        assert beta == 0.0, "beta must be 0.0 if out is None"
    else:
        out = np.asarray(out)
        if out.shape != c_shape:
            msg = f"Output shape {out.shape} does not match expected shape {c_shape} for subscripts '{subscripts}'"
            raise ValueError(msg)

    # handle zero-sized input
    # if A or B is size zero, mult quits early and does not scale C.
    if is_trivial:
        shift(out, subscript_c, alpha=0.0, beta=beta)
    else:
        mult(a, b, out, subscript_a, subscript_b, subscript_c, alpha=alpha, beta=beta, conja=conja, conjb=conjb)
    return out


def contract(
    subscripts: str,
    a: npt.ArrayLike,
    b: npt.ArrayLike,
    alpha: scalar = 1.0,
    beta: scalar = 0.0,
    out: Optional[npt.ArrayLike] = None,
    conja: bool = False,
    conjb: bool = False,
    allow_partial_trace: bool = False,
    complex_real_contractions: bool = False,
) -> npt.ArrayLike:
    """
    Similar to `pytblis.contract`, but with support for contractions between complex and real tensors.
    If one of the input tensors is complex and the other is real, we perform separate contractions for the
    real and imaginary parts of the complex tensor.
    C (stored in `out` if provided) is computed as:
    C = alpha * einsum(subscripts, a, b) + beta * C if `out` is provided.

    Parameters
    ----------
    subscripts : str
        Subscripts defining the contraction.
    a : array_like
        First tensor operand.
    b : array_like
        Second tensor operand.
    alpha : scalar, optional
        Scaling factor for the product of `a` and `b`.
    beta : scalar, optional
        Scaling factor for the output tensor. Must be 0.0 if `out` is None.
    conja: bool, optional
        If True, conjugate the first tensor `a` before contraction. Alpha is not conjugated.
    conjb: bool, optional
        If True, conjugate the second tensor `b` before contraction. Beta is not conjugated.
    allow_partial_trace : bool, optional
        If True, handle redundant indices in subscripts for `a` and `b` by doing partial trace before contraction.
    complex_real_contractions : bool, default False
        If True, handle contractions between complex and real tensors by performing
        separate contractions for the real and imaginary parts of the complex tensor.
        alpha and beta must be real in this case.
    out : array_like, optional
        Output tensor to store the result.

    Returns
    -------
    ndarray
        Result of the tensor contraction.

    Examples
    --------
    >>> import numpy as np
    >>> from pytblis import contract_complex_real
    >>> A = np.random.rand(3, 4, 5)
    >>> B = np.random.rand(4, 5, 6).astype(np.complex128)
    >>> result = contract_complex_real('ijk,jkl->il', A, B, alpha=2.0)
    """

    if not complex_real_contractions:
        return contract_same_type(
            subscripts,
            a,
            b,
            alpha=alpha,
            beta=beta,
            out=out,
            conja=conja,
            conjb=conjb,
            allow_partial_trace=allow_partial_trace,
        )
    a = np.asarray(a)
    b = np.asarray(b)

    if out is not None:
        real_scalar_type = _check_tblis_types(a.real, b.real, out=out.real)
    else:
        real_scalar_type = _check_tblis_types(a.real, b.real)

    if real_scalar_type is None:
        raise TypeError("Input arrays and output array (if provided) must use the same IEEE floating point type.")

    if np.iscomplexobj(a) and np.iscomplexobj(b):
        return contract(subscripts, a, b, alpha, beta, out, conja, conjb, allow_partial_trace)
    if (not np.iscomplexobj(a)) and (not np.iscomplexobj(b)):
        return contract(subscripts, a, b, alpha, beta, out, conja, conjb, allow_partial_trace)

    # exactly one of a or b is complex, so the result must be complex
    if out is not None and not np.iscomplexobj(out):
        msg = "Output array must have complex dtype when contracting a complex tensor with a real tensor."
        raise ValueError(msg)

    result_type = np.result_type(real_scalar_type, 1j).type
    # allocate complex output if not provided

    c_shape = contraction_result_shape(subscripts, a.shape, b.shape)
    if out is None:
        out = np.empty(c_shape, dtype=result_type)
        assert beta == 0.0, "beta must be 0.0 if out is None"

    assert alpha.imag == 0.0, "alpha must be real when contracting a complex tensor with a real tensor."
    assert beta.imag == 0.0, "beta must be real when contracting a complex tensor with a real tensor."

    if c_shape:
        if np.iscomplexobj(a) and not np.iscomplexobj(b):
            imag_fac = -1.0 if conja else 1.0
            a_real = a.real
            contract(
                subscripts,
                a_real,
                b,
                alpha=alpha,
                beta=beta,
                out=out.real,
                conja=conja,
                conjb=conjb,
                allow_partial_trace=allow_partial_trace,
            )
            a_imag = a.imag
            contract(
                subscripts,
                a_imag,
                b,
                alpha=alpha * imag_fac,
                beta=beta,
                out=out.imag,
                conja=conja,
                conjb=conjb,
                allow_partial_trace=allow_partial_trace,
            )
        else:
            imag_fac = -1.0 if conjb else 1.0
            b_real = b.real
            contract(
                subscripts,
                a,
                b_real,
                alpha=alpha,
                beta=beta,
                out=out.real,
                conja=conja,
                conjb=conjb,
                allow_partial_trace=allow_partial_trace,
            )
            b_imag = b.imag
            contract(
                subscripts,
                a,
                b_imag,
                alpha=alpha * imag_fac,
                beta=beta,
                out=out.imag,
                conja=conja,
                conjb=conjb,
                allow_partial_trace=allow_partial_trace,
            )
    # handle scalar output
    # c_shape == ()
    elif np.iscomplexobj(a) and not np.iscomplexobj(b):
        imag_fac = -1.0 if conja else 1.0
        out = (
            beta * out
            + alpha * contract(subscripts, a.real, b, conja=conja, conjb=conjb, allow_partial_trace=allow_partial_trace)
            + 1j
            * alpha
            * imag_fac
            * contract(subscripts, a.imag, b, conja=conja, conjb=conjb, allow_partial_trace=allow_partial_trace)
        )
    else:
        imag_fac = -1.0 if conjb else 1.0
        out = (
            beta * out
            + alpha * contract(subscripts, a, b.real, conja=conja, conjb=conjb, allow_partial_trace=allow_partial_trace)
            + 1j
            * alpha
            * imag_fac
            * contract(subscripts, a, b.imag, conja=conja, conjb=conjb, allow_partial_trace=allow_partial_trace)
        )
    return out


def ascontiguousarray(a):
    """Parallel transpose the input to C-contiguous layout.

    Parameters
    ----------
    a : array_like
        Input array.

    Returns
    -------
    ndarray
        Contiguous array.
    """
    a = np.asarray(a)
    if a.flags.c_contiguous:
        return a

    if not _check_strides(a):
        warnings.warn(
            f"Input tensor of shape {a.shape} has non-positive strides: {a.strides}. Falling back to numpy ascontiguousarray.",
            stacklevel=2,
        )
        return np.ascontiguousarray(a)

    if a.dtype.type not in _accepted_types:
        warnings.warn(
            "TBLIS only supports float32, float64, complex64, and complex128. Falling back to numpy ascontiguousarray.",
            stacklevel=2,
        )
        return np.ascontiguousarray(a)

    out = np.empty(a.shape, dtype=a.dtype, order="C")
    assert len(a.shape) < len(_valid_labels), (
        f"a.ndim is {len(a.shape)}, but only {len(_valid_labels)} labels are valid."
    )
    a_inds = _valid_labels[: len(a.shape)]
    a_inds = "".join(a_inds)
    add(a, out, a_inds, a_inds, alpha=1.0, beta=0.0)
    return out


def asfortranarray(a):
    """Parallel transpose the input to F-contiguous layout.

    Parameters
    ----------
    a : array_like
        Input array.

    Returns
    -------
    ndarray
        Fortran-order array.
    """
    a = np.asarray(a)
    if a.flags.f_contiguous:
        return a

    if not _check_strides(a):
        warnings.warn(
            f"Input tensor of shape {a.shape} has non-positive strides: {a.strides}. Falling back to numpy asfortranarray.",
            stacklevel=2,
        )
        return np.asfortranarray(a)

    if a.dtype.type not in _accepted_types:
        warnings.warn(
            "TBLIS only supports float32, float64, complex64, and complex128. Falling back to numpy asfortranarray.",
            stacklevel=2,
        )
        return np.asfortranarray(a)

    out = np.empty(a.shape, dtype=a.dtype, order="F")
    assert len(a.shape) < len(_valid_labels), (
        f"a.ndim is {len(a.shape)}, but only {len(_valid_labels)} labels are valid."
    )
    a_inds = _valid_labels[: len(a.shape)]
    a_inds = "".join(a_inds)
    add(a, out, a_inds, a_inds, alpha=1.0, beta=0.0)
    return out
